package ru.dbotthepony.mc.otm.core.collect

import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import it.unimi.dsi.fastutil.objects.Object2ObjectRBTreeMap
import it.unimi.dsi.fastutil.objects.ObjectIterators
import net.minecraft.core.BlockPos
import net.minecraft.core.SectionPos
import net.minecraft.world.level.ChunkPos
import ru.dbotthepony.kommons.collect.iterateSetBits
import java.util.BitSet
import java.util.Collections

class BlockPosSet : MutableSet<BlockPos> {
	private class Segment {
		val bits = BitSet(16 * 16 * 16)

		private fun pos2index(pos: BlockPos): Int {
			val x = pos.x.and(15)
			val z = pos.z.and(15) shl 4
			val y = pos.y.and(15) shl 8
			return x or y or z
		}

		private fun index2pos(index: Int, gX: Int, gY: Int, gZ: Int): BlockPos {
			val x = index.and(15) + gX
			val z = index.ushr(4).and(15) + gZ
			val y = index.ushr(8).and(15) + gY
			return BlockPos(x, y, z)
		}

		operator fun contains(pos: BlockPos): Boolean {
			return bits[pos2index(pos)]
		}

		fun add(pos: BlockPos): Boolean {
			val index = pos2index(pos)

			if (bits[index])
				return false

			bits[index] = true
			return true
		}

		fun remove(pos: BlockPos): Boolean {
			val index = pos2index(pos)

			if (!bits[index])
				return false

			bits[index] = false
			return true
		}

		fun iterator(gX: Int, gY: Int, gZ: Int): MutableIterator<BlockPos> {
			return object : MutableIterator<BlockPos> {
				private val iterator = bits.iterateSetBits(4096)
				private var last = -1

				override fun hasNext(): Boolean {
					return iterator.hasNext()
				}

				override fun next(): BlockPos {
					last = iterator.nextInt()
					return index2pos(last, gX, gY, gZ)
				}

				override fun remove() {
					if (last == -1)
						throw NoSuchElementException()

					bits[last] = false
					last = -1
				}
			}
		}

		val isEmpty: Boolean
			get() = bits.isEmpty

		val size: Int
			get() = bits.cardinality()
	}

	private val segments = Object2ObjectRBTreeMap<SectionPos, Segment>(Vec3iHashStrategy)

	fun subset(pos: ChunkPos): Set<BlockPos> {
		val result = BlockPosSet()

		for (key in segments.keys.iterator(SectionPos.of(pos.x, 0, pos.z))) {
			if (key.x != pos.x || key.z != pos.z) break
			result.segments[key] = segments[key]
		}

		return Collections.unmodifiableSet(result)
	}

	override fun add(element: BlockPos): Boolean {
		return segments.computeIfAbsent(SectionPos.of(element), Object2ObjectFunction { Segment() }).add(element)
	}

	override val size: Int
		get() = segments.values.sumOf { it.size }

	override fun addAll(elements: Collection<BlockPos>): Boolean {
		var any = false
		elements.forEach { any = add(it) || any }
		return any
	}

	override fun clear() {
		segments.clear()
	}

	override fun contains(element: BlockPos): Boolean {
		return segments[SectionPos.of(element)]?.contains(element) == true
	}

	override fun containsAll(elements: Collection<BlockPos>): Boolean {
		return elements.all { contains(it) }
	}

	override fun isEmpty(): Boolean {
		var any = false

		segments.entries.removeIf { (_, s) ->
			any = any || s.isEmpty
			s.isEmpty
		}

		return any
	}

	override fun iterator(): MutableIterator<BlockPos> {
		return object : MutableIterator<BlockPos> {
			private val iterator = segments.entries.iterator()
			private var current: MutableIterator<BlockPos> = ObjectIterators.emptyIterator()
			private var last = current
			private var foundNext = false

			private fun findNext() {
				if (!foundNext) {
					foundNext = true

					while (!current.hasNext() && iterator.hasNext()) {
						val (pos, itr) = iterator.next()
						current = itr.iterator(pos.x shl 4, pos.y shl 4, pos.z shl 4)
					}
				}
			}

			override fun hasNext(): Boolean {
				findNext()
				return current.hasNext()
			}

			override fun next(): BlockPos {
				findNext()
				last = current
				foundNext = false
				val blockPos = current.next()
				return blockPos
			}

			override fun remove() {
				last.remove()
			}
		}
	}

	override fun remove(element: BlockPos): Boolean {
		val index = SectionPos.of(element)
		val segment = segments[index] ?: return false

		if (segment.remove(element)) {
			if (segment.isEmpty)
				segments.remove(index)

			return true
		}

		return false
	}

	override fun removeAll(elements: Collection<BlockPos>): Boolean {
		var any = false
		elements.forEach { any = remove(it) || any }
		return any
	}

	override fun retainAll(elements: Collection<BlockPos>): Boolean {
		return removeIf { it !in elements }
	}
}
